function [FinalX, fv, gfv, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, eigHess] = LyapunovLR(A, M, LM, B, b, n, p, ManiParams, SolverParams, HasHHR, initialX, preconditioner)
% This code solves the following optimization problem: 
%        min f([Y]) := trace(Y^T * A * Y * Y^T * M * Y + Y^T * M * Y * Y^T * A * Y )
%                      - trace(Y^T * B * Y)
%        s.t. [Y] \in \mathbb{R}^{n \times p}_* / \mathcal{O}_p
%
% Input:
%   A : matrix, SPD, n x n
%   M : matrix, SPD, n x n
%   LM: matrix, LM = chol(M, 'lower') or []
%   B : matrix, b * b^T or B when b = [], n x n
%   b : matrix, low-rank matrix such that B = b * b^T when b is not [], n x s

    if isempty(B) && isempty(b)
        error('B and b all are zero matrix!')
    end

    % compute only once the perfect shuffle matrix PI.
    if preconditioner > 0 && preconditioner < 5
        PI = perfectshuffle(p, p);
    else
        preconditioner = 0;
        PI = [];
        disp('Set preconditioner = 0, as it is not one of {1,2,3,4}.')
    end    
    
    global numSystemsb numSystemsB timeOfSystems;
    numSystemsb = 0;   
    numSystemsB = 0;   
    timeOfSystems = 0;           

    % problem-related handle
    fhandle = @(x)f(x, A, M, B, b, n, p);
    gfhandle = @(x)gf(x, A, M, B, b, n, p);
    Hesshandle = @(x, eta)Hess(x, eta, A, M, B, b, n, p);    
    PreConhandle = @(x, eta)PreCon(x, eta, A, M, B, b, preconditioner, LM, PI);
    
    [FinalX, fv, gfv, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, eigHess] = ...
        DriverOPT(fhandle, gfhandle, Hesshandle, PreConhandle, SolverParams, ManiParams, HasHHR, initialX);          
    FinalX.numSystemsb = numSystemsb;
    FinalX.numSystemsB = numSystemsB;
    FinalX.timeOfSystems = timeOfSystems;
    
    clear numSystems

end

% computing perfect shuffle matrix
function L = perfectshuffle(n, p)
    L = sparse(n * p, n * p);
    for h = 1 : p
       for k = 1 : n
           i = (k - 1) * p + h;
           j = (h - 1)* n + k;
           L(i, j) = 1;
       end
    end
end

% computing functional value
function [output, x] = f(x, A, M, B, b, n, p)
% f(Y) = trace(Y^T * A * Y * Y^T * M * Y + Y^T * M * Y * Y^T * A * Y) - trace(Y^T * C * Y)
    Y = x.main;
    x.AY = A * Y;
    x.YAY = Y' * x.AY;
    x.MY = M * Y;
    x.YMY = Y' * x.MY;    

    if isempty(B)
        x.BY = b * (b' * Y); 
    else
        x.BY = B * Y;
    end
    % tips: trace(A^T * B) = sum(sum(A .* B))
    output = sum(sum(x.YAY .* x.YMY)) - sum(sum(x.main .* x.BY));      
end

% computing the Euclidean gradient of the cost function
function [output, x] = gf(x, A, M, B, b, n, p)
% gradf_Y = A * Y * Y^T * M * Y + M * Y * Y^T * A * Y - B * Y
    output.main = 2 * (x.AY * x.YMY + x.MY * x.YAY - x.BY);    
end

% computing the action of the Euclidean Hessian to eta
function [output, x] = Hess(x, eta, A, M, B, b, n, p)
%%/*exixTV = 2 * A * eta * Y^T * M * Y + 2 * M * eta * Y^T * A * Y + 2 * A * Y * eta^T * M * Y + 2 * M * Y * eta^T * A * Y + 2 * A * Y * Y^T * M * eta + 2 * M * Y * Y^T * A * eta - 2 * C * eta */

%     Y = x.main;
%     Meta = M * eta.main;
%     output.main = 2 * ((A * eta.main) * x.YMY + Meta * x.YAY + x.AY * (eta.main' * x.MY) + x.MY * (eta.main' * x.AY) + x.AY * (Y' * Meta) + x.MY * (x.AY' * eta.main) - (b * (b' * eta.main)));    

    etaMY = eta.main' * x.MY;  % eta^T * M * Y
    etaAY = eta.main' * x.AY;  % eta^T * M * Y
    if isempty(B)
        output.main = b * (b' * eta.main);
    else
        output.main = B * eta.main;
    end
    output.main = 2 * ((A * eta.main) * x.YMY + (M * eta.main) * x.YAY + x.AY * (etaMY + etaMY') + x.MY * (etaAY + etaAY') - output.main);
end

% preconditioning on A and M
function [output, x] = PreCon(x, eta, A, MM, B, b, preconditioner, LM, PI)
% Solving the following problem 
%      Approx(Hess(Y)[xi]) = eta for xi
% 
% preconditioner: integer
%   0   : no precontioner
%   1   : new preconditioer under the first metric
%   2   : new preconditioer under the first metric
%   3   : new preconditioer under the first metric
%   4   : preconditioner in Bart's paper but MM is the identity matrix

    if preconditioner == 0 % no preconditioner
        output = eta;           
        return;

%% preconditioning under first metric         
    elseif preconditioner == 1 % new perconditioner under the first metric                   
        [n, p] = size(eta.main);        
%         numSystems = numSystems + p;
        Y = x.main;
        if isempty(LM)   % O(p^3)
            L = chol(x.YMY, 'lower');  % may be not numberically stable 
        else
            [~, L] = qr0(LM' * Y); % O(n^2p+np^2)
            L = L';
        end

%         LY = chol(YY, 'lower');  % may be not numberically stable
        YY = Y' * Y;    % O(n^2p)
        [~, LY] = qr0(Y);   % O(np^2)
        LY = LY';    

        LYAYL = L \ x.YAY / L';  % O(np^2)
        LYAYL = 0.5 * (LYAYL + LYAYL');  % O(p^2)
        [Q, Lambda] = eigs(full(LYAYL),p);    % O(p^3)
%         Q = real(Q); Lambda = real(Lambda);

        LQ = L' \ Q;        % O(p^2)
        Yeta = Y' * eta.main;  % O(np^2)
        QLYA = (LQ)' * (x.AY)'; % O(np^2)               % Q^T * L^{-1} * Y^T * A 
        [hatV, ~] = qr(full(x.MY), 0); % O(np^2)        % hatV = orth(M * Y)
                 % O(np^2+np+p^3)
        tmp0 = eta.main + Y * (LY' \ (LY \ (Yeta)));    % double, tmp0 = (I + P_Y) * eta                  
                 % O(np^2+np+p^3)
        tmp1 = mat2cell((tmp0 - hatV * (hatV' * tmp0)) * ((Y' * Y) * (LQ)), ...
                         [n], ones(1, p));              % cell,   tmp1 = (I - hatV * hatV^T) * (I + P_Y) * eta * (Y^T * Y) * L^{-T} * Q                       
                 % O(np)
        tmp2 = 2 * QLYA';                               % double, tmp2 = 2 * A * Y * L^{-T} * Q
                 % O(np^2+np)
        tmp2 = tmp2 - hatV * (hatV' * tmp2);            % double, tmp2 = (I - hatV * hatV^T) * 2 * A * Y * L^{-T} * Q         
                
        cLambda = num2cell(diag(Lambda)');   % cell         
                % O(p(np^c+np^2+np))        
        TMP = cellfun(@(x, y) subproblem1AM(A, MM, x, hatV, y, tmp2), cLambda, tmp1, 'UniformOutput', false);
                % O(np^2(p+1))
        U   = cell2mat(cellfun(@(x) QLYA * x(:,1), TMP, 'UniformOutput', false));
                % O(np^3(p+1))
        K   = cellfun(@(x, y) (2 * x) * speye(p) - QLYA * y(:, 2:end), cLambda, TMP, 'UniformOutput', false);            
                % O(p^2)
        tSxi = subproblem2(K, 2 * (LQ)' * (Yeta) * (YY) * (LQ), U, PI);                
        ctSxi = mat2cell(tSxi, [p], ones(1,p));
                % O()
        tZxi  = cell2mat(cellfun(@(x, y) (x(:, 1) - x(:, 2 : end) * y), TMP, ctSxi, 'UniformOutput', false));        
                % O(p^3)
        Sxi = LQ * tSxi * LQ';
                % O(np^3)
        Zxi = tZxi * LQ';
                % O(np^2)
        tmp = LY' \ (LY \ (Y' * Zxi));
                % O(np^2+np)
        Zxi1 = 0.5 * (Y * (tmp + tmp'));
                % O(np^2 + np)
        Zxi2 = Zxi - Y * tmp;
                % O(np^2+np)
        output.main = Y * Sxi + Zxi2 + Zxi1;

%% preconditioning under second metric 
    elseif preconditioner == 2 % new preconditioner under the second metric                
        [n, p] = size(eta.main);   
%         numSystems = numSystems + p;
        Y = x.main;
        if isempty(LM)
            L = chol(x.YMY, 'lower');  % may be not numberically stable
        else
            [~, L] = qr0(LM' * Y);
            L = L';
        end

        YY = Y' * Y;   
        [~, LY] = qr0(Y);
        LY = LY';

        LYAYL = L \ x.YAY / L';
        LYAYL = 0.5 * (LYAYL + LYAYL');
        [Q, Lambda] = eig(full(LYAYL));
%         Q = real(Q); Lambda = real(Lambda);

        LQ = L' \ Q;
        Yeta = Y' * eta.main;
        QLYA = (LQ)' * (x.AY)';              % Q^T * L^{-1} * Y^T * A 
        hatV = qr(full(x.MY), 0);            % hatV = orth( M * Y)
        tmp0 = 0.5 * eta.main * (YY * LQ);   % tmp0 = 0.5 * eta * (Y^T * Y) * L^{-T} * Q
        tmp1 = mat2cell(tmp0 - hatV * (hatV' * tmp0), ...
                        [n], ones(1, p));    % tmp1 = (I - hatV * hatV^T) * 0.5 * eta * (Y^T * Y) * L^{-T} * Q
        tmp2 = 2 * QLYA';                    % tmp2 = 2 * A * Y * L^{-T} * Q
        tmp2 = tmp2 - hatV * (hatV' * tmp2); % tmp2 = (I - hatV * hatV^T) * 2 * A * Y * L^{-1} * Q        
        
        cLambda = num2cell(diag(Lambda)');        
        TMP = cellfun(@(x, y) subproblem1AM(A, MM, x, hatV, y, tmp2), cLambda, tmp1, 'UniformOutput', false);
        U   = cell2mat(cellfun(@(x) QLYA * x(:,1), TMP, 'UniformOutput', false));
        K   = cellfun(@(x, y) (2 * x) * speye(p) - QLYA * y(:, 2:end), cLambda, TMP, 'UniformOutput', false);               

        tSxi = subproblem2(K, 0.5 * (LQ)' * (Yeta) * (YY) * (LQ), U, PI);
        ctSxi = mat2cell(tSxi, [p], ones(1,p));
        tZxi  = cell2mat(cellfun(@(x, y) (x(:, 1) - x(:, 2 : end) * y), TMP, ctSxi, 'UniformOutput', false)); 

        Sxi = (LQ) * tSxi * (LQ)';
        Zxi = tZxi * (LQ)';

        tmp = (LY' \ (LY \ (Y' * Zxi)));
        Zxi1 = 0.5 * (Y * (tmp + tmp'));
        Zxi2 = Zxi - Y * tmp;
        output.main = Y * Sxi + Zxi2 + Zxi1;
 
%% preconditioning under third metric 
    elseif preconditioner == 3 % new preconditioner under the third metric                
        [n, p] = size(eta.main);        
%         numSystems = numSystems + p;
        Y = x.main;
        if isempty(LM)
            L = chol(x.YMY, 'lower');  % may be not numberically stable
        else
            [~, L] = qr0(LM' * Y);
            L = L';
        end
        
        [~, LY] = qr0(Y); 
        LY = LY';         % YY^T = LY * LY^T
        YY = LY * LY';

        LYAYL = L \ full(x.YAY) / L';
        LYAYL = 0.5 * (LYAYL + LYAYL');
        [Q, Lambda] = eig(LYAYL);
%         Q = real(Q); Lambda = real(Lambda);

        LQ = L' \ Q;                
        QLYA = (LQ)' * (x.AY)';              % Q^T * L^{-1} * Y^T * A 
        hatV = qr(full(x.MY), 0);            % hatV = orth(MM * Y)
        tmp0 = 0.5 * eta.main * LQ;          % tmp0 = 0.5 * eta * L^{-T} * Q
        tmp1 = mat2cell(tmp0 - hatV * (hatV' * tmp0), ...
                        [n], ones(1, p));    % tmp1 = (I - hatV * hatV^T) * 0.5 * eta * L^{-T} * Q
        tmp2 = 2 * QLYA';                    % tmp2 = 2 * A * Y * L^{-T} * Q
        tmp2 = tmp2 - hatV * (hatV' * tmp2); % tmp2 = (I - hatV * hatV^T) * 2 * A * Y * L^{-1} * Q        

        cLambda = num2cell(diag(Lambda)');                  
        TMP = cellfun(@(x, y) subproblem1AM(A, MM, x, hatV, y, tmp2), cLambda, tmp1, 'UniformOutput', false);            
        U   = cell2mat(cellfun(@(x) QLYA * x(:,1), TMP, 'UniformOutput', false));
        K   = cellfun(@(x, y) (2 * x) * speye(p) - QLYA * y(:, 2:end), cLambda, TMP, 'UniformOutput', false);            

        hSxi = subproblem2(K, 0.5 * (LQ)' * (Y' * eta.main) * (LQ), U, PI);
        chSxi = mat2cell(hSxi, [p], ones(1, p));
        tZxi  = cell2mat(cellfun(@(x, y) (x(:, 1) - x(:, 2 : end) * y), TMP, chSxi, 'UniformOutput', false)); 

        tSxi = (LQ) * hSxi * (LQ)';
        Zxi = tZxi * (LQ)';
        % solving for Sxi by Sxi * (Y^TY) + (Y^TY) * Sxi = 2 * (Y^TY) * tSxi * (Y^TY)        
        Z = 2 * YY * tSxi * YY;
%         Sxi = subproblem3(Z, P, Gamma);
        Sxi = lyap(YY, - Z);                
        Omega = lyap(YY,  Zxi' * Y - Y' * Zxi);
        output.main = Y * (LY' \ (LY \ Sxi)) + Zxi - Y * Omega;            

%% preconditioning under the first metric in Bart's papar 
    elseif preconditioner == 4 % preconditioner in Bart's paper        
        [n, p] = size(eta.main);        
%         numSystems = numSystems + p;
        Y = x.main;
        [V, R] = qr(Y, 0);
        tmp  = eta.main * R';
        Seta = 2. * V' * tmp;
        Zeta = tmp - 0.5 * V * Seta;
        VAV  = V' * (A * V);
        VAV  = 0.5 * (VAV + VAV');
        [Q, Lambda] = eig(full(VAV));        
        tV = V * Q;
        AtV = A * tV;
        tmp1 = AtV - tV * (tV' * AtV);
        tSeta = Q' * Seta * Q;
        
        tZeta = mat2cell(Zeta * Q, [n], ones(1,p));
        cLambda = num2cell(diag(Lambda)');
        
        TMP = cellfun(@(x, y) subproblem1(A, x, tV, y, tmp1), cLambda, tZeta, 'UniformOutput', false);
        U   = cell2mat(cellfun(@(x) AtV' * x(:, 1), TMP, 'UniformOutput', false));
        K   = cellfun(@(x, y) x * speye(p) - AtV' * y(:, 2 : end), cLambda, TMP, 'UniformOutput', false);        

%         for i = 1 : p
%             tmp2(:,:,i) = subproblem1(A, Lambda(i, i), tV, [tZeta(:, i), AtV - tV * (tV' * AtV)]);
%             U(:, i) = AtV' * tmp2(:, 1, i);
%             K{i} = Lambda(i, i) * speye(p) - AtV' * tmp2(:, 2 : end, i);
%         end

        tSxi = subproblem2(K, tSeta, U, PI);
        ctSxi = mat2cell(tSxi, [p], ones(1, p));
        tZxi  = cell2mat(cellfun(@(x, y) x(:, 1) - x(:, 2 : end) * y, TMP, ctSxi, 'UniformOutput', false));

        Sxi = Q * tSxi * Q';
        Zxi = tZxi * Q';        
        output.main = (0.5 * V * Sxi + Zxi) / (R');
    end
end


function output = invALtimesM(A, MM, lambda_i, B)  % O(np^c)
global numSystemsb numSystemsB timeOfSystems
numSystemsB = numSystemsB + 1;
numSystemsb = numSystemsb + size(B,2);
tt = tic;
    ALtimesM = A + lambda_i * MM; 
    output = ALtimesM \ B;                     % ZWH, TODO: implement efficient computing inverse of a matrix
%     output = cholmod2(ALtimesM, B, -1);   % sometimes go wrong  
timeOfSystems = timeOfSystems + toc(tt);
end


function output = subproblem1(A, lambda_i, tV, B1, B2)
% Solving the saddle-point problems
    n = size(tV,1);
    B = [B1, B2];
    tmp = invALtimesM(A, speye(n), lambda_i, [B, tV]);    
    Si = tV' * tmp(:, size(B, 2) + 1 : end);
    N = Si \ (tV' * tmp(:, 1 : size(B, 2)));
    output = tmp(:, 1 : size(B, 2)) - tmp(:, size(B, 2) + 1 : end) * N;
end

function output = subproblem1AM(A, MM, lambda_i, tV, B1, B2)  % O(np^c+np^2+np)
% Solving the saddle-point problems   
    B = [B1, B2];
    tmp = invALtimesM(A, MM, lambda_i, [B, tV]); % O(np^c)
    % (A + lambda * M)^{-1} * B
    AplamMB = tmp(:, 1 : size(B, 2)); 
    % (A + lambda * M)^{-1} * tV
    AplamMtV = tmp(:, size(B, 2) + 1 : end); 
    Si = tV' * AplamMtV;        % O(np^2)
    N = Si \ (tV' * AplamMB);          % O(np^2)
    output = AplamMB - AplamMtV * N;   % O(np^2+np)
end

function output = subproblem1WithAM(A, MM, lambda_i, tV, B, invALMtV) 
% B = (I - hatV * hatV^T) * 2 * A * Y * L^{-1} * Q
% output = (A + lambda_i * M)^{-1} * B - (A + lambda_i * M)^{-1} * hatV * S_i^{-1} * (hatV + (A + lambda_i * M)^{-1} * B)    
    tmp = invALtimesM(A, MM, lambda_i, B);
    Si = tV' * invALMtV;
    N = Si \ (tV' * tmp);
    output = tmp - invALMtV * N;
end

function output = subproblem2(K, tSeta, U, PI)  % O(p^2)
% Solving the linear system:  (K + \Pi * K * \Pi) vec(\tilde{S}_\xi) = vec{R}    
    p = size(tSeta, 1);
    function output = LinearOper(S)
        S = reshape(S, p, p);
        tmp1 = zeros(p, p);
        tmp2 = zeros(p, p);
        for i = 1 : p
            tmp1(:, i) = K{i} * S(:, i);
            tmp2(i, :) = S(i, :) * K{i}';
        end
        output = reshape(tmp1 + tmp2, [], 1);
    end

    KK = sparse(p * p, p * p);
    for j = 1 : p
        KK((j - 1) * p + 1 : j * p, (j - 1) * p + 1 : j * p) = K{j};
    end

    L = ichol(sparse(KK + PI * KK * PI));
    b = tSeta - U - U';
    [output, ~] = pcg(@LinearOper, b(:), 1e-12, p * p, L, L');
    output = reshape(output, p, p);
    
%     b = tSeta - U - U';
%     output = reshape(myCG(@LinearOper, b(:), 1e-12, p * p), p, p);
    
end

function output = subproblem3(Z, U, Lambda)
% Solving for X by X * E + E * X = Z
% Used in the new preconditioner under the third metric.
% Input: E = U * Lambda * U^T, eigenvalue decomposition
    p = size(Z,1);        
    UTZU = U' * Z * U;
    UTXU = zeros(p,p);
    for i = 1:p
        for j = 1:p
            UTXU(i,j) = UTZU(i,j) / (Lambda(i,i) + Lambda(j,j));
        end
    end
    output = U * UTXU * U';
end

function output = myCG(Axhandle, b, tol, maxiter)
    x = zeros(size(b));
    r = -b;
    p = -r;
    k = 0;
    while(norm(r) > tol && k < maxiter)
        Ap = Axhandle(p);
        alpha = - r' * p / (p' * Ap);
        x = x + alpha * p;
        Ax = Axhandle(x);
        r = Ax - b;
        beta = r' * Ap / (p' * Ap);

        p = -r + beta * p;

        k = k + 1;
    end
    output = x;
end

function [Q, R] = qr0(A)
% qr decomposition for A = Q * R, where Q^T * Q = I, R is upper triangular and
% its diagonal elements are not negative.
    [Q, R] = qr(full(A), 0);
    tmp = sign(diag(R));    
    P = diag(tmp);
    R = P * R;
    Q = Q * P;
end
