function [FinalY, outInfo, rel_res] = RLyapSolver(A, M, LM, B, b, ExtParams, IntParams)
%   This function handle aims to compute a low-rank approximation to the following generalized Lyapunov equation 
%                           AXM + MXA = C, 
% where A, M are S.P.D., and C is S.P.S.D. 
% Note that, we work on the factor Y satisfying X = YY^T.
%
% Input: 
%   A : matrix, SPD, n x n
%   M : matrix, SPD, n x n
%   LM: matrix, LM = chol(M, 'lower') or [],
%   B : matrix, b * b^T or [] when b = [], n x n
%   b : matrix, low-rank matrix such that B = b * b^T when b is not [], n x s
%   ExtParams: struct including 
%     p_Min  : integer,  the initial rank approximation the solution X
%     p_Inc  : integer,  the increament of rank after each outer iteration
%     p_Max  : integer,  the maximum rank approximation for the solution X
%     numLs  : integer,  the number of linesearch in outer iteration
%     ext_Tol: double, the tolerance
%   IntParams: struct including 
%     ParamSet      : integer, one in [1, 2, 3, 4, 5] for different Riemannian matrics
%     methods       : specify the Riemannian method for Problem (3.2)
%     Min_Iteration : minimum number of iterations for solving Problem (3.2)
%     Max_Iteration : maximum number of iterations for solving Problem (3.2)
%     Stop_Criterion: 
%     OutputGap     :
%     Verbose       :
%     Precond       :
%     Int_Tol       : double, 
% Output:
%   outInfo: struct including
%     iter: number of iterations
%     nf  : number of evaluation of cost function
%     ng  : number of evaluation of gradient
%     nR  : number of calling Retraction
%     nV  : number of first call to vector transport
%     nVp : number of call to vector transport
%     nH  : number of the action of Hessian 
%     ComTime: totally CPU time 
%     funs   : function value set of each iteration
%     grads  : norm of gradient set of each iteration 
%     times  : cumulative time of each iteration
%   rel_res  : relative residual array of each iteration

    tic
    n = size(A, 1);    
    l = size(b, 2);    
    ManiParams.name = 'SymFixedRankQ';
    ManiParams.n = n;
    ManiParams.ParamSet = IntParams.ParamSet;
    IntParams.Max_Inner_Iter = 50;       

    normb = norm(b' * b, 'fro');
    iter_ = 0; nf_ = 0; ng_ = 0; nR_ = 0; nV_ = 0; nVp_ = 0; nH_ = 0; totalTime = 0;
    numSystemsb = 0;
    numSystemsB = 0;
    timeOfSystems = 0;
    HasHHR = 0;
    funs_ = [];
    grads_ = [];
    times_ = [];


    if ~isfield(ExtParams,'initialY')
%         scalar = 2 * norm(A, 'fro') * norm(M, 'fro');
%         scalar =  norm(b * b','fro') / sqrt(scalar);
%         initialY.main = scalar * orth(randn(n, ExtParams.p_Min));     
        nY = rand(n, ExtParams.p_Min);
%         initialY.main = (scalar / norm(nY)) * nY;    
        initialY.main = nY / norm(nY);    
    else
        initialY.main = ExtParams.initialY;
    end
    isTol = false;
    if ~isfield(IntParams, 'Tolerance')
        isTol = true;
        IntParams.Tolerance = 1e-1;
    end
    IntParams.Accuracy = 1e-8;

    rel_res = [];
    totalTime = totalTime + toc;
    for p = ExtParams.p_Min : ExtParams.p_Inc : ExtParams.p_Max
        tic
        ManiParams.p = p;
        totalTime = totalTime + toc;
        [FinalY, fv, gfv, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, ~, ~, ~, ~] =...
                LyapunovLR(A, M, LM, B, b, n, p, ManiParams, IntParams, HasHHR, initialY, IntParams.precond);
        tic
        totalTime = totalTime + ComTime;        
        funs_ = [funs_; fv];
        grads_ = [grads_; gfv];
        iter_ = iter_ + iter; nf_ = nf_ + nf; ng_ = ng_ + ng; nR_ = nR_ + nR; nV_ = nV_ + nV; nVp_ = nVp_ + nVp; nH_ = nH_ + nH;          
        numSystemsb = numSystemsb + FinalY.numSystemsb;
        numSystemsB = numSystemsB + FinalY.numSystemsB;
        timeOfSystems = timeOfSystems + FinalY.timeOfSystems;
        % fastly computing relative residual
        if isempty(B)
            FF = full([FinalY.AY, FinalY.MY, b]);
            [~, R] = qr(FF, 0);         
            T = [zeros(p, p), eye(p), zeros(p, l); eye(p), zeros(p, p), zeros(p, l); zeros(l, p), zeros(l, p), -eye(l)];        
            residual = norm(R * T * R', 'fro') / normb;        
            rel_res = [rel_res;residual];
        else
            % TODO
        end       

        % checking convergence
        if(residual <= ExtParams.ext_Tol)
            totalTime = totalTime + toc;
            times_ = [times_; totalTime];
            break
        end

        if isTol
            IntParams.Tolerance = 1e-2 / 5;
        end
        
        % getting descent direction
        [V, ~] = eigs(@(v)Tx(v, FinalY, A, M, B, b, n, p), n, ExtParams.p_Inc);
        Y.main = [FinalY.main, real(V)];
        if IntParams.ParamSet == 1 || IntParams.ParamSet == 4
            [~, RY] = qr(Y.main, 0);
            Ddir.main = ((A * FinalY.main) * (FinalY.main' * (M * Y.main)) + M * FinalY.main * (FinalY.main' * (A * Y.main)) - b * (b' * Y.main)) / RY / RY';
            Ddir.main = Ddir.main - 0.5 * (Y.main / RY / RY') * (Y.main' * Ddir.main);            
        elseif IntParams.ParamSet == 2 || IntParams.ParamSet == 5
            [~, RY] = qr(Y.main, 0);
            Ddir.main = 2 * (A * FinalY.main * (FinalY.main' * (M * Y.main)) + M * FinalY.main * (FinalY.main' * (A * Y.main)) - b * (b' * Y.main)) / RY / RY';            
        elseif IntParams.ParamSet == 3 || IntParams.ParamSet == 6
            Ddir.main = 2 * (A * FinalY.main * (FinalY.main' * (M * Y.main)) + M * FinalY.main * (FinalY.main' * (A * Y.main)) - b * (b' * Y.main));
        end
        Ddir.main = Ddir.main / norm(Ddir.main, 'fro');
        
        % increasing rank
        Y.main = [FinalY.main, zeros(n, ExtParams.p_Inc)];        
        [fYv, Y] = f(Y, A, M, B, b, n, p);  nf_ = nf_ + 1;
        [gfY, Y] = gf(Y, A, M, B, b, n, p); ng_ = ng_ + 1;
        
        % performing one step descent direction
        alpha = 1;
        Ynew.main = Y.main - Ddir.main;   %  Xnew.main = X.main - alpha * Ddir.main
        [fYnewv, Ynew] = f(Ynew, A, M, B, b, n, p); nf_ = nf_ + 1;
        ExtParams.numLs = 20; nls = 0;
%         while (nls < ExtParams.numLs && fXv - fXnewv < 1e-4 * alpha * norm(gfX.main, 'fro')^2 && abs(fXv - fXnewv) > abs(fXv) * 1e-10)
        while (nls < ExtParams.numLs && fYnewv > fYv + 1e-4 * alpha * sum(sum(gfY.main .* Ddir.main)) && abs(fYv - fYnewv) > abs(fYv) * 1e-10)
            alpha = alpha * 0.5;
            Ynew.main = Y.main - alpha * Ddir.main;
            [fYnewv, Ynew] = f(Ynew, A, M, B, b, n, p);            
            nls = nls + 1;
        end
        nf_ = nf_ + nls;
        [gfY, Ynew] = gf(Ynew, A, M, B, b, n, p); ng_ = ng_ + 1;
        gfY =  EucGradtoGrad(gfY.main, Ynew, IntParams.ParamSet);
        ngf0 = metric(gfY, gfY, Ynew.main, IntParams.ParamSet);
        ngf0 = sqrt(ngf0);        
                
        % adaptly adjusting tolerance          
%         exp1 = getExp(ngf0);
%         if(residual > 1e-6)
%             exp1 = 7 + exp1;  % 7, gf / gf0 < 1e-6
%         else
%             exp1 = 12 + exp1; % 12, gf / gf0 < 1e-11
%         end
%         IntParams.Tolerance = min(10^(-exp1),1e-2);  

%         if (IntParams.method == "LRBFGS" || IntParams.method == "RTRNewton") && ( residual <= 5.e-6)
%             IntParams.Accuracy = 1e-8;
%         else
%             IntParams.Accuracy = 1e-8;            
%         end
        
        initialY.main = Ynew.main;        
        totalTime = totalTime + toc;
        fprintf('\n>>n:%d,rank:%d,res:%.3e(%.1e),nls:%d,fv:%.3e, ngf0:%.3e,inner_tol:%.1e,time:%.3e\n', ...
                 n, p, residual, ExtParams.ext_Tol, nls, fYnewv, ngf0, IntParams.Tolerance, totalTime);        
        times_ = [times_; totalTime];
    end
    fprintf('\n>>>n:%d,rank:%d,res:%.3e(%.1e),fv:%.3e, gf:%.3e,time:%.3e\n', ...
                 n, p, residual, ExtParams.ext_Tol, fv, gfv,totalTime);    
    outInfo.iter = iter_; 
    outInfo.nf = nf_;
    outInfo.ng = ng_; 
    outInfo.nR = nR_; 
    outInfo.nV = nV_; 
    outInfo.nVp = nVp_; 
    outInfo.nH = nH_; 
    outInfo.ComTime = totalTime;
    outInfo.funs = funs_;
    outInfo.grads = grads_;
    outInfo.times = times_;
    outInfo.numSystemsb = numSystemsb;
    outInfo.numSystemsB = numSystemsB;
    outInfo.timeOfSystems = timeOfSystems;
end

function output =  metric(u, v, Y, option)
    if option == 1 || option == 4
        output = 2 * (sum(sum((Y' * u)' .* (Y' * v))) + sum(sum((Y' * Y) .* (u' * v))));
    elseif option == 2 || option == 5
        output = sum(sum(((Y' * Y) .* (u' * v))));
    elseif option == 3 || option == 6
        output = sum(sum((u .* v)));
    end    
end

function output =  EucGradtoGrad(grad, x, option)
    if option == 1 || option == 4 
        [~, R] = qr(x.main, 0);
        output = 0.5 * grad / R / R';
        output = output - 0.5 * (x.main / R / R') * (x.main' * output);
    elseif option == 2 || option == 5
        [~, R] = qr(x.main, 0);
        output = grad / R / R';
    elseif option == 3 || option == 6
        output = grad;
    end
end

function output = Tx(v, X, A, M, B, b, n, p)
    Y = X.main;
    [Q, ~] = qr(Y, 0);
    tmp = v - Q * (Q' * v);
    if isempty(B)
        Btmp = b * (b' * tmp); 
    else
        Btmp = B * tmp;
    end 
    tmp = X.AY * (X.MY' * tmp) + X.MY * (X.AY' * tmp) - Btmp;
    output = tmp - Q * (Q' * tmp);
end

% computing functional value
function [output, x] = f(x, A, M, B, b, n, p)
    Y = x.main;
    x.AY = A * Y;
    x.YAY = Y' * x.AY;
    x.MY = M * Y;
    x.YMY = Y' * x.MY;
    if isempty(B)
        x.BY = b * (b' * Y); 
    else
        x.BY = B * Y;
    end
    % tips: trace(A^T * B) = sum(sum(A .* B))
    output = sum(sum(x.YAY .* x.YMY)) - sum(sum(x.main .* x.BY));      
end

% computing the Euclidean gradient of the cost function
function [output, x] = gf(x, A, M, B, b, n, p)
%     Y = x.main;
    output.main = 2 * (x.AY * x.YMY + x.MY * x.YAY - x.BY);    
end

function  output = getExp(number)   
% get the exponent of a float number, e.g., getExp(123.213) = 2 as 123.213 = 1.23213e2
    output = 0;
    while(number < 1.)
        number = number * 10;
        output = output - 1;
    end
    while(number >= 10.)
        number = number * 0.1;
        output = output + 1;
    end
end
