function [FinalX, fv, gfv, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, eigHess] = LyapunovLRE(A, M, LM, B, b, n, p, ManiParams, SolverParams, HasHHR, initialX, preconditioner)
% This function solves the following optimization problem: 
%        min f(X) := trace(AXMX +  MXAX) - trace(BX)
%        s.t. X \in S_+(p,n) = {X in R^{nxn}:X=X^T, X>=0, rank(X)=p},
%        where X >= 0 means that X is symmetric positive semidefinite.
%
% The implementation follows [VV10, Section 6.3]. The iterate consists of
% two matrices V and D, and they are stored in an array by column, i.e.,
% iterate = [V(:); D(:)].
%
% [VV10]: Bart Vandereycken and Stefan Vandewalle. A Riemannian optimization approach for computing low-rank solutions of Lyapunov equations. SIAM Journal on Matrix Analysis and Applications, 31(5):2553–2579, 2010.
%
% Input:
%   A : matrix, SPD, n x n
%   M : matrix, SPD, n x n
%   LM: matrix, LM = chol(M, 'lower') or []
%   B : matrix, b * b^T or B when b = [], n x n
%   b : matrix, low-rank matrix such that B = b * b^T when b is not [], n x s

    if isempty(B) && isempty(b)
        error('B and b all are zero matrix!')
    end

    % compute only once the perfect shuffle matrix PI.
    if preconditioner > 0 && preconditioner < 5
        PI = perfectshuffle(p, p);
    else
        preconditioner = 0;
        PI = [];
        disp('Set preconditioner = 0, as it is not one of {1,2,3,4}.')
    end    
    
    global numSystems timeOfSystems
    numSystems = 0;   
    timeOfSystems = 0;

    % problem-related handle
    fhandle      = @(x)f(x, A, M, LM, B, b, n, p);
    gfhandle     = @(x)gf(x, A, M, B, b, n, p);
    Hesshandle   = @(x, eta)Hess(x, eta, A, M, B, b, n, p);    
    PreConhandle = @(x, eta)PreCon(x, eta, A, M, LM, B, b, n, p, PI, preconditioner);
    
    [FinalX, fv, gfv, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, eigHess] = ...
        DriverOPT(fhandle, gfhandle, Hesshandle, PreConhandle, SolverParams, ManiParams, HasHHR, initialX);          
    FinalX.numSystems = numSystems;
    FinalX.timeOfSystems = timeOfSystems;
    
    clear numSystems timeOfSystems
end


%%  Problem related function handles
% computing functional value
function [output, x] = f(x, A, M, LM, B, b, n, p)
    V = reshape(x.main(1:n*p), n, p);
    D = reshape(x.main(n*p+1:end), p, p);
    
    x.V = V;
    x.D = D;
    x.AV = A * V;
    x.MV = M * V; 
    x.VAV = V' * x.AV;

    if isempty(LM)
        x.VMV = V' * x.MV;       
    else
        [~, R] = qr(LM' * V, 0);
        x.VMV = R' * R;
    end    

    if isempty(B)
        x.BV = b * (b' * V); 
    else
        x.BV = B * V;
    end
    % tips: trace(A^T * B) = sum(sum(A .* B))
    output = trace( (x.VAV * D) * (x.VMV * D) ) - trace(V' * (x.BV * D));       
end

% computing the Riemannian of the cost function
function [output, x] = gf(x, A, M, B, b, n, p)       
    T = (x.AV * x.D) * x.VMV + (x.MV * x.D) * x.VAV - x.BV;    
    S = sym(x.V' * T);
    Z = T - x.V * S;
    output.main = [Z(:); S(:)];    
end

% computing the action of the Riemannian Hessian to eta
function [output, x] = Hess(x, eta, A, M, B, b, n, p)    
    Z = reshape(eta.main(1:n*p), n, p);
    S = reshape(eta.main(n*p+1:end), p, p);

    ZMV = Z'*x.MV;
    ZAV = Z'*x.AV;
    T1 = x.AV*S*x.VMV+x.MV*S*x.VAV + ...
         x.AV*(ZMV) + x.MV*(ZAV) + ...
         (A*Z)*x.VMV + (M*Z)*x.VAV;
    T2 = (x.AV * x.D) * (ZMV') + (x.MV * x.D) * (ZAV') - b*(b'*Z);
    
    SS = sym(x.V'*T1);
    ZZ = T1 - x.V*(x.V'*T1) + (T2 - x.V*(x.V'*T2)) / x.D;
    output.main = [ZZ(:); SS(:)];    
end


function output = proj(x, eta, n, p)  % there may be ways to be more efficient 
    V = reshape(x.main(1:n*p), n, p);    

    Z = reshape(eta.main(1:n*p), n, p);
    S = reshape(eta.main(n*p+1:end), p, p);

    VZ = V' * Z;
    S = sym(S + VZ + VZ');
    Z = Z - x.V * VZ;
    output.main = [Z(:); S(:)];
end


% preconditioning on A and M
function [output, x] = PreCon(x, eta, A, MM, LM, B, b, n, p, PI, pre)
% Solving the following problem 
%      Approx(Hess(Y)[xi]) = eta for xi
% 
% preconditioner: integer
%   0   : no precontioner
%   1   : the proposed preconditioer
%   2   : the preconditioner in Bart's paper

    if pre == 0     % no preconditioner
        output = eta;
    elseif pre == 1 % new perconditioner under the first metric         
        output = PreCon_1(x, eta, A, MM, LM, B, b, n, p, PI);
    elseif pre == 2   % preconditioner in Bart's paper;                     
        output = PreCon_Bart(x, eta, A, MM, LM, B, b, n, p, PI);
    else
        error('error: this option does not exist for precond!')
    end
end

function [output, x] = PreCon_1(x, eta, A, MM, LM, B, b, n, p, PI)
    Z = reshape(eta.main(1:n*p), n, p);
    S = reshape(eta.main(n*p+1:end), p, p);        
    
    rr = sqrt(diag(x.D));   % rr = diag(R) = sqrt(sqrt(D))  
%     R = sqrtm(x.D);
    Y = x.V .* rr';         % Y = V*R. Note that A * diat(x) = A .* x', where x is a col vector
    if isempty(LM)
        YMY = r .* (x.VMV .* rr');   
        L = chol(sym(YMY), 'lower');  % may be not numberically stable 
    else
        [~, L] = qr0(LM' * Y); % O(n^2p+np^2)
        L = L';
    end

    etamain = (0.5 * x.V * S + Z) ./ rr';    

%     LY = chol(YY, 'lower');  % may be not numberically stable
    [~, LY] = qr0(Y);  % O(np^2)
    LY = LY';    
    
    LAV = chol(full(sym(x.VAV)),"lower");
    [~, LV] = qr(LAV' .* rr', 0);    
    LYAYL = sym(L \ (LV' * (LV / L')));  % L \ ((R * (x.VAV * R)) / L');  % O(np^2)
    [Q, Lambda] = eig(full(LYAYL));  % O(p^3)
    
    Yeta = Y' * etamain;  % O(np^2)
    QLYA = Q' * (L \ (x.AV .* rr')');            % O(np^2)     Q^T * L^{-1} * Y^T * A 
    [hatV, ~] = qr(full(x.MV .* rr'), 0);     % O(np^2)     hatV = orth(M * Y)
    tmp0 = etamain + Y * (LY' \ (LY \ (Yeta)));    % double, tmp0 = (I + P_Y) * eta                  
    tmp1 = mat2cell((tmp0 - hatV * (hatV' * tmp0)) * (LY * ((LY' / L') * Q)), ...
                     [n], ones(1, p));             % cell,   tmp1 = (I - hatV * hatV^T) * (I + P_Y) * eta * (Y^T * Y) * L^{-T} * Q                       
    tmp2 = 2 * QLYA';                              % double, tmp2 = 2 * A * Y * L^{-T} * Q
    tmp2 = tmp2 - hatV * (hatV' * tmp2);           % double, tmp2 = (I - hatV * hatV^T) * 2 * A * Y * L^{-T} * Q         
    
    cLambda = num2cell(diag(Lambda)');   % cell                     
    TMP = cellfun(@(x, y) subproblem1AM(A, MM, x, hatV, y, tmp2), cLambda, tmp1, 'UniformOutput', false);    
    U   = cell2mat(cellfun(@(x) QLYA * x(:,1), TMP, 'UniformOutput', false));            
    K   = cellfun(@(x, y) (2 * x) * speye(p) - QLYA * y(:, 2:end), cLambda, TMP, 'UniformOutput', false);            

    tSxi = subproblem2(K, 2 * Q' * (L \ Yeta) * (LY * ((LY' / L')*Q)), U, PI);
    tSxi = sym(tSxi);
    ctSxi = mat2cell(tSxi, [p], ones(1,p));

    tZxi  = cell2mat(cellfun(@(x, y) (x(:, 1) - x(:, 2 : end) * y), TMP, ctSxi, 'UniformOutput', false));        
 
    Sxi = L' \ (Q * ((tSxi * Q') / L));  % Sxi = L \ (Q * ((tSxi * Q') / L'));           
    Zxi = (tZxi * Q') / L;               % Zxi = tZxi * Q^T * L^-1            
    tmp = LY' \ (LY \ (Y' * Zxi));            
    Zxi1 = Y * sym(tmp);            
    Zxi2 = Zxi - Y * tmp;
    
    tmp = (Y * Sxi + Zxi1 + Zxi2) .* rr'; 
    S = 2. * (x.V' * tmp);
    S  = sym(S); 
    Z = tmp - 0.5 * x.V * S;

    output.main = [Z(:); S(:)]; 
    output = proj(x, output, n, p);
end

function [output, x] = PreCon_Bart(x, eta, A, MM, LM, B, b, n, p, PI)
    Zeta = reshape(eta.main(1:n*p), n, p);
    Seta = reshape(eta.main(n*p+1:end), p, p);                      
%         VAV  = V' * (A * V);
    [Q, Lambda] = eig(full(sym(x.VAV)));        
    tV = x.V * Q;
    AtV = A * tV;
    tmp1 = AtV - tV * (tV' * AtV);
    tSeta = Q' * (Seta * Q);
    
    tZeta = mat2cell(Zeta * Q, [n], ones(1,p));
    cLambda = num2cell(diag(Lambda)');
    
    TMP = cellfun(@(x, y) subproblem1(A, x, tV, y, tmp1), cLambda, tZeta, 'UniformOutput', false);
    U   = cell2mat(cellfun(@(x) AtV' * x(:, 1), TMP, 'UniformOutput', false));
    K   = cellfun(@(x, y) x * speye(p) - AtV' * y(:, 2 : end), cLambda, TMP, 'UniformOutput', false);        

    tSxi = subproblem2(K, tSeta, U, PI);
    tSxi = sym(tSxi);
    ctSxi = mat2cell(tSxi, [p], ones(1, p));
    tZxi  = cell2mat(cellfun(@(x, y) x(:, 1) - x(:, 2 : end) * y, TMP, ctSxi, 'UniformOutput', false));

    Sxi = Q * tSxi * Q';
    Zxi = tZxi * Q';

    output.main = [Zxi(:); Sxi(:)];
    output = proj(x, output, n, p);
end

%% other useful tools
function X = sym(X)
    X = 0.5 * (X + X');
end

function [Q, R] = qr0(A)
% qr decomposition for A = Q * R, where Q^T * Q = I, R is upper triangular and
% its diagonal elements are not negative.
    [Q, R] = qr(A, 0);
    tmp = diag(R);
    tmp = sign(tmp);
    P = diag(tmp);
    R = P * R;
    Q = Q * P;
end

function output = invALtimesM(A, MM, lambda_i, B)  % O(np^c)
global numSystems timeOfSystems
numSystems = numSystems + size(B,2);
tt = tic;
    ALtimesM = A + lambda_i * MM; 
    output = ALtimesM \ B;                     % ZWH, TODO: implement efficient computing inverse of a matrix
%     output = cholmod2(ALtimesM, B, -1);   % sometimes go wrong  
timeOfSystems = timeOfSystems + toc(tt);
end


function output = subproblem1(A, lambda_i, tV, B1, B2)
% Solving the saddle-point problems
    n = size(tV,1);
    B = [B1, B2];
    tmp = invALtimesM(A, speye(n), lambda_i, [B, tV]);    
    Si = tV' * tmp(:, size(B, 2) + 1 : end);
    N = Si \ (tV' * tmp(:, 1 : size(B, 2)));
    output = tmp(:, 1 : size(B, 2)) - tmp(:, size(B, 2) + 1 : end) * N;
end

function output = subproblem1AM(A, MM, lambda_i, tV, B1, B2)  % O(np^c+np^2+np)
% Solving the saddle-point problems   
    B = [B1, B2];
    tmp = invALtimesM(A, MM, lambda_i, [B, tV]); % O(np^c)
    % (A + lambda * M)^{-1} * B
    AplamMB = tmp(:, 1 : size(B, 2)); 
    % (A + lambda * M)^{-1} * tV
    AplamMtV = tmp(:, size(B, 2) + 1 : end); 
    Si = tV' * AplamMtV;        % O(np^2)
    N = Si \ (tV' * AplamMB);          % O(np^2)
    output = AplamMB - AplamMtV * N;   % O(np^2+np)
end

function output = subproblem2(K, tSeta, U, PI)  % O(p^2)
% Solving the linear system:  (K + \Pi * K * \Pi) vec(\tilde{S}_\xi) = vec{R}    
    p = size(tSeta, 1);
    function output = LinearOper(S)
        S = reshape(S, p, p);
        tmp1 = zeros(p, p);
        tmp2 = zeros(p, p);
        for i = 1 : p
            tmp1(:, i) = K{i} * S(:, i);
            tmp2(i, :) = S(i, :) * K{i}';
        end
        output = reshape(tmp1 + tmp2, [], 1);
    end

    KK = sparse(p * p, p * p);
    for j = 1 : p
        KK((j - 1) * p + 1 : j * p, (j - 1) * p + 1 : j * p) = K{j};
    end

    L = ichol(sparse(KK + PI * KK * PI));
    b = tSeta - U - U';
    [output, ~] = pcg(@LinearOper, b(:), 1e-10, p * p, L, L');
    output = reshape(output, p, p);
    
%     b = tSeta - U - U';
%     output = reshape(myCG(@LinearOper, b(:), 1e-12, p * p), p, p);
end

% computing perfect shuffle matrix
function L = perfectshuffle(n, p)
    L = sparse(n * p, n * p);
    for h = 1 : p
       for k = 1 : n
           i = (k - 1) * p + h;
           j = (h - 1)* n + k;
           L(i, j) = 1;
       end
    end
end